#############################################################################################################
#  
#   Detecting Audience Costs Replication
#
#   Note: You must save the lib_cPBE folder in your directory.
#
#   Taehee Whang
#   08/20/2014  
#	
#############################################################################################################

###########
## Model ##
###########

rm(list = ls())

# You must change the working directory
setwd("C:/Users/TW/Dropbox/Kurizaki_prj/dac/IO_RnR/Acceptance/replication_materials")
libname="lib_cPBE/Release/lib_cPBE.dll"
#libname="lib_cPBE/Release/lib_cPBE.so"

LinuxCluster = 0
# if 1, snow library is used.
# if 0, "fake" snow library is used so that snow function works without real cluster.

# calculate equilibrium probability - a vectorized version.
get_pr_vec <-function(psets_t,tol=.Machine$double.eps^0.5){
  psets = t(psets_t)
  n = nrow(psets_t)
  pfrc = .C("solve_eq_cpbe__",
      as.integer(n),
          as.double(psets),
      as.double(tol),
          pfrc = double(n*4))$pfrc
  pfrc = matrix(pfrc,nrow=4,ncol=n)
  pfrc = t(pfrc)
  colnames(pfrc) <- c('pf','pr','pc','pf_prior')
  return(pfrc)
}

# calculate equilibrium probability - pset is assumed to be an array.
get_pr <- function(pset,tol=.Machine$double.eps^0.5){
    return(get_pr_vec(t(pset))[1,])
}

################
## Estimation ##
################

estimate <-function(pset0, method="Nelder-Mead") {

    optim_result = optim(pset0, likelihood_model_snow,
                method = method,   # "Nelder-Mead", "BFGS", "CG", "L-BFGS-B", "SANN"
                control = list(trace=0,maxit=10^5) # ,reltol=1e-10)
                )
    pset_hat = optim_result$par
    names(pset_hat) <- g_reg_pset_names
    return(pset_hat)
}

# likelihood is actually log-likelihood
likelihood_model <- function(reg_pset,dataf)
  sum(likelihood_individual(reg_pset,dataf))

# likelihood of each observation
likelihood_individual <-function(reg_pset,dataf){
    psets = make_pset(reg_pset,dataf)  # calculate expected values for each node.
    lik = matrix(0, nrow=nrow(dataf), ncol=1)
    lik_penalty = matrix(g_penalty , nrow=nrow(dataf), ncol=1)

    pfrc = get_pr_vec(psets) # equilibrium probabilities

    if (pfrc[1,1]< 0) return(lik_penalty)

    psq = 1 - pfrc[,3]
    pcd = pfrc[,3] * (1-pfrc[,2])
    pbd = pfrc[,3] * pfrc[,2] * (1-pfrc[,1])
    psf = pfrc[,3] * pfrc[,2] * pfrc[,1]

        z=1:nrow(dataf)
        lik[dataf$y_sq * z] = log(psq[dataf$y_sq * z])
        lik[dataf$y_cd * z] = log(pcd[dataf$y_cd * z])
        lik[dataf$y_bd * z] = log(pbd[dataf$y_bd * z])
        lik[dataf$y_sf * z] = log(psf[dataf$y_sf * z])
    # lik = log(psq)*dataf$y_sq + log(pcd)*dataf$y_cd + log(pbd)*dataf$y_bd + log(psf)*dataf$y_sf
    # log-likelihood
    if (all(is.finite(lik)))
        return(-lik)
    else
        return(lik_penalty)
}

likelihood_at_each_node <- function(p) {
      ll = likelihood_model(p,g_datf_sub) #g_datf_sub should be assigned at each node before using this function
      return(sum(unlist(ll)))
    }

# likelihood using snow
likelihood_model_snow <- function(reg_pset){
  lik_list = clusterCall(cl,likelihood_at_each_node,reg_pset)
  if (all(lik_list<g_penalty*nrow(g_datf)/(length(cl)+1))) {
    return( sum(unlist(lik_list)))
  } else {
    return( g_penalty * nrow(g_datf) )
  }
}

# calculate likelihood for each observation, using snow
likelihood_individual_snow <- function(reg_pset){
  lik_list = clusterCall(cl,
    function(p) likelihood_individual(p,g_datf_sub),reg_pset)
  return( unlist(lik_list))
}

# make model parameters, based on regressors and given coef.
# using global variables, g_reg_names and g_pset_names
make_pset <- function(reg_pset,dataf){
    psets = matrix(0,nrow = nrow(dataf), ncol = length(g_pset_names) )
    colnames(psets) <- g_pset_names
    for (i in 1:length(g_pset_names)){
        for (j in 1: (length(g_reg_names)/2)  ){
            if (g_pset_names[i]==g_reg_names[j*2-1]) {
                psets[,i] = psets[,i] + data.matrix(dataf[g_reg_names[j*2]]) * reg_pset[j]
                        }
        }
    }
    # do not estimate sVWB and sVVB
    psets[,'sVVB']=0
    psets[,'sVWB']=0

    return(psets)
}

# variance estimates, based on BHHH(outer product method)
get_asymptiotic_var <- function(lnlik,x,n_obs,dif = 10^-14 ){
# lnlik : a function of a parameter that returns a column vector of log likelihood for each observation.
# x : MLE parameter
  G = matrix(0,nrow = n_obs, ncol = length(x) )
  for (i in 1:length(x)){
    x0 = x
    x0[i] = x0[i] - dif/2
    x1 = x
    x1[i] = x1[i] + dif/2
    y0 = lnlik(x0)
    y1 = lnlik(x1)
    G[,i] = (y1 - y0)/ dif
  }
  BHHH = solve(t(G) %*% G )
  return(BHHH)
}

#make each parameters name.
get_reg_pset_names <- function(reg_names){
    reg_pset_names = rep('',times=length(reg_names)/2)
    for (i in 1:(length(reg_names)/2))
        reg_pset_names[i] = paste(reg_names[i*2-1],reg_names[i*2],sep="_")
    return(reg_pset_names)
}

  
########################
##   INITIALIZATION   ##
########################

ncl = 5 # number of clusters. When using fake version, it will be changed to 1.
if (LinuxCluster==1)
    {
    library(snow)
    } else  # snow function is fake now.
    {
        ncl = 1  # there is no actual cluster.
        makeCluster <- function(ncl){
            return(list(1))   # only one node
        }
        
        clusterCall <- function(cl,fun,...){
            temp = do.call(fun,list(...))
            return(list(temp))
        }

        clusterEvalQ <- function(cl,fun){
            temp = eval(fun)
            return(list(temp))
        }

        clusterSetupSPRNG <- function(cl, seed){
            set.seed(seed)
        }

        clusterApply <- function(cl,X, FUN, ...){
            lapply(X, FUN, ...)
        }

        clusterSplit <- function(cl,x) { list(x) }

        clusterExport <- function(cl,li){ }

    }


cl = makeCluster(ncl)

library(foreign)
library(nlme)
temp = clusterEvalQ(cl,library(MASS))

di<-dyn.load(libname)
temp = clusterCall(cl,"dyn.load",libname)

# stopCluster(cl)
.C("init_cpbe_dll__")    # initialize dll
temp = clusterCall(cl,".C","init_cpbe_dll__")


####################################
##       Regressor setting        ##
####################################

g_pset = 0   # for debugging

g_penalty = 100 # if 0 probability is assigned, this value is given as a penalty log-likelihood
g_pset_names = c    ('mSA','mVA','maA','mWA','sVVA','sVaA','sVWA','saWA','sWWA','mCB','mVB','mWB','sVVB','sVWB') # used in make_pset. Do not change the order.
pset1 =     c     (    0,    1, -1/2,   -1,     0,     0,     0,     0,     1,    0,    1,   -1,     0,     0)
names(pset1) <-g_pset_names

# regressor names - this consists of pairs of (nodename,regressor).
# nodename - one of g_pset_names
# regressor - one of column names in data.frame

# ACQ of State A: ACQ_A
mvar2 = c( 'mVA','const', 'mVA','tau_regi', 'mVA','civilwar_B', 'mVA','contig' ) 			
ini_mvar2 = c( 1.77, -1.18, 1.07, 0.72 )    #'mVA','const','tau_regi','civilwar_B','contig'

# BD of State A: BD_A
maar1 = c( 'maA','const', 'maA','dem_A' )
ini_maar1 = c(-4.06, -0.76) #'maA','const','dem_A',

# SF of State A: SF_A   
mwar1 = c( 'mWA','const', 'mWA','capshare_A', 'mWA','dem_A', 'mWA','lnencap_A' )
ini_mwar1 = c( -3.62, 0.83, -0.79, 0.017    ) #'mWA','const','capshare_A','dem_A','lnencap_A',

# ACQ of State B: ACQ_B 
mcbr1 = c( 'mCB','const', 'mCB','tau_regi', 'mCB','civilwar_B', 'mCB','contig' )
ini_mcbr1 = c( -1.41, 0.19, 0.15, -0.091 ) #'mCB','const','tau_regi','civilwar_B','contig'

# SF of State B: SF_B   
mwbr1 = c( 'mWB','const', 'mWB','capshare_A', 'mWB','dem_B', 'mWB','lnencap_B' )
ini_mwbr1 = c( -1.97, 1.17, 0, -0.021 ) #'mWB','const','capshare_A','dem_B','lnencap_B' 

# Var-Cov initial values
ini_vcv = c( 1.241,-0.23, 0.23, -0.26, 1.145)

####################
##     Data       ##
####################

g_datf = read.dta("dac_IO_main.dta",convert.underscore=FALSE)
g_datf['const']=1  # add a constant column
g_datf['zero']=0  # add a zero column
attach(g_datf)

temp=clusterExport(cl,ls()) # prepare for snow

g_sub_list = clusterSplit(cl,1:nrow(g_datf)) # list of sequences, for each node.
g_datf_sub_list = lapply(g_sub_list, function(i) g_datf[i,]) # list of sub_data. Each node will process one sub_data.
temp=clusterApply(cl,g_datf_sub_list,
  function(x) assign("g_datf_sub",x,envir = .GlobalEnv)) # Each node gets its own sub_data as a global variable.

print(unlist(clusterCall(cl,function() nrow(g_datf_sub)))) # number of observations that each node will process.

make_pfrc <- function(reg_pset,dataf){
    psets = make_pset(reg_pset,dataf)  # calculate expected values for each node.
    pfrc = get_pr_vec(psets) # equilibrium probabilities
    #save(pfrc,file="pfrc.Rdata")
    return(pfrc)
}


#############
# BEGIN RUN #
#############

## Main model replication

g_reg_names = c(mvar2, maar1, mwar1, mcbr1, mwbr1, 'sVVA','const', 'sVaA','const', 'sVWA','const', 'saWA','const', 'sWWA','const')
g_reg_pset_names = get_reg_pset_names(g_reg_names) # for display later
reg_pset_ini = c(ini_mvar2, ini_maar1, ini_mwar1, ini_mcbr1, ini_mwbr1, ini_vcv ) 

ptm=proc.time()
reg_pset2 = estimate(reg_pset_ini,"BFGS")
reg_pset3 = estimate(reg_pset2)
reg_pset_KW = estimate(reg_pset3,"BFGS")

aV=get_asymptiotic_var(likelihood_individual_snow,reg_pset_KW,nrow(g_datf))
reg_pset_KW_sd = diag(aV) ^ 0.5
est_summary = cbind(as.matrix(reg_pset_KW),reg_pset_KW_sd, as.matrix(reg_pset_KW/reg_pset_KW_sd) )
est_summary
pfrc = make_pfrc(reg_pset_KW,g_datf)
log_likelihood = likelihood_model(reg_pset_KW,g_datf)

mle_KW = list(est_summary=est_summary, reg_pset_KW=reg_pset_KW, aV=aV, pfrc=pfrc, initial_value=reg_pset_ini, log_likelihood=log_likelihood)
print((proc.time()-ptm)[3])
mle_KW$est_summary
mle_KW$log_likelihood


###########
# END RUN #
###########


#RESULTS:

#> mle_KW$est_summary
#                            reg_pset_KW_sd           
#mVA_const      -1.472518411    1.112204183 -1.3239641
#mVA_tau_regi   -2.518984026    1.367247673 -1.8423758
#mVA_civilwar_B  4.067382668    2.130853886  1.9088041
#mVA_contig      1.128500864    0.777759777  1.4509633
#maA_const      -5.975798310    1.568245733 -3.8104987
#maA_dem_A      -0.323809113    0.096430896 -3.3579395
#mWA_const      -3.331099568    1.244939338 -2.6757124
#mWA_capshare_A  1.297127058    0.801233520  1.6189126
#mWA_dem_A      -0.094428606    0.042788552 -2.2068661
#mWA_lnencap_A   0.099395362    0.059157068  1.6801942
#mCB_const      -0.397836864    0.394125467 -1.0094168
#mCB_tau_regi    0.666771051    0.354048265  1.8832773
#mCB_civilwar_B -0.425573666    0.371299102 -1.1461748
#mCB_contig     -0.173355995    0.262605462 -0.6601386
#mWB_const      -1.058994812    0.385332489 -2.7482625
#mWB_capshare_A  0.501315869    0.341621650  1.4674593
#mWB_dem_B       0.010181978    0.008367975  1.2167792
#mWB_lnencap_B  -0.004731784    0.018384014 -0.2573858
#sVVA_const      2.722771420    1.827919086  1.4895470
#sVaA_const      0.578018137    1.555184751  0.3716717
#sVWA_const     -0.680565375    0.837899506 -0.8122279
#saWA_const     -0.682081667    1.115711500 -0.6113423
#sWWA_const      1.295768198    0.839393386  1.5436960

#> mle_KW$log_likelihood
#[1] 368.8941
